;-----------------------------------------------------------------------------
;
; Paul Houle (paulhoule.com) Created $$20171110 -- code moved out of
; par2asm.cpp to build native 64-bit version, as 64-bit CL doesn't allow
; inline assembly.
;
; This is part of par2cmdline (a PAR 2.0 compatible file verification and
; repair tool). See http://parchive.sourceforge.net for details of PAR 2.0.
;
; par2cmdline is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; par2cmdline is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;-----------------------------------------------------------------------------

	;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; The .asm code shares one source between 32 and 64 bit versions,
	; with mode-specific differences handled by the if64() macro.
	; This "single source" restriction constrains the 64-bit code (for
	; instance, r8-r15 effectively can't be utilized).
	;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	.nolist
	include	par2x86.inc
	if64(<;>, <.686>)
	if64(<;>, <.mmx>)
	if64(<;>, <.model flat, C>)
	.code
	.list

;-----------------------------------------------------------------------------
;
; Executes single emms instruction (harmless if not using mms opcodes).
;
;-----------------------------------------------------------------------------

Asm_emms proc
;??? Not sure if doing this in 64-bit mode is a problem or not, since MS
;??? Visual C compiler requires mm6 and mm7 must preserved in 64-bit builds.
;??? Might be best to disallow any use of mmx opcodes in 64-bit builds,
;??? replacing with, for instance, SSE?
	emms
	ret
Asm_emms endp

;-----------------------------------------------------------------------------
;
; Return processor capabilities as a dword -- used by SetPr().
;
;-----------------------------------------------------------------------------

Asm_getpr proc	uses xbx
	xpushf				;try to flip "ID" bit in eflags
	pop     xax
	btc     eax,21
	push    xax
	xpopf
	xpushf
	pop	xcx
	xor     ecx,eax			;bit 21 == 0 if "ID" bit changed
	xor	eax,eax
	bt      ecx,21
	jc	fset			;jmp= no CPUID, no advanced opcodes
	cpuid				;eax= highest supported CPUID function
	test	eax,eax
	jz	fset			;jmp= can't get features
	push	1			;get features
	pop	xax
	cpuid
	xor	eax,eax
	test	dl,1			;floating point?
	jz	fset			;jmp= no, no MMX
	bt	edx,23			;MMX support?
	jnc	fset			;jmp= nope
	or	eax,GETPR_MMX		;indicate MMX
fset:
	inc	eax			;insure return value is non-zero
	ret
Asm_getpr endp

;-----------------------------------------------------------------------------
;
; Builds the combined multiplication tables at the specified address (pMtab).
;
; Note: apparently in Win 64 mm6/mm7 must be saved across func calls -- this
; was news to me.  Previous use of mm7 in this routine was changed to the
; (thankfully available) mm5 to avoid this issue.
;
;-----------------------------------------------------------------------------

Asm_BuildMtab proc uses xbx xsi xdi, \
			pMtab:ptr dword, factor:dword, table:ptr void

	push	if64(xcx, pMtab)	;(move via push/pop for 64-bit case)
	mov	eax,if64(edx, factor)
	movzx	ebx,ah
	movzx	eax,al
	shl	ebx,8
	shl	eax,8

	mov	xdx,if64(r8, table)
	lea	xsi,[xdx+xax*2]			;LL
	lea	xcx,[xdx+xbx*2+(1*256*256)*2]	;HL
	lea	xdi,[xdx+xax*2+(1*256*256)*2]	;LH
	lea	xdx,[xdx+xbx*2+(2*256*256)*2]	;HH

	xor	ebx,ebx
	pxor	mm5,mm5
	pop	xax
	align	16
LHmxlp:
	movq	mm0,[xsi+xbx]
	pxor	mm0,[xcx+xbx]		;L3,L2,L1,L0
	movq	mm1,mm0
	punpcklwd mm0,mm5
	punpckhwd mm1,mm5
	movq	[xax+xbx*2],mm0
	movq	[xax+xbx*2+8],mm1

	movq	mm0,[xsi+xbx]
	pxor	mm0,[xcx+xbx]		;L3,L2,L1,L0
	movq	mm1,mm0
	punpcklwd mm0,mm5
	punpckhwd mm1,mm5
	movq	[xax+xbx*2],mm0
	movq	[xax+xbx*2+8],mm1

	movq	mm2,[xdi+xbx]
	pxor	mm2,[xdx+xbx]		;H3,H2,H1,H0
	movq	mm3,mm2
	punpcklwd mm2,mm5
	punpckhwd mm3,mm5
	movq	[xax+400h+xbx*2],mm2
	movq	[xax+400h+xbx*2+8],mm3

	add	ebx,8
	cmp	ebx,512
	jne	LHmxlp

	ret
Asm_BuildMtab endp

;-----------------------------------------------------------------------------
;
; Processes block of data a multiple of 4 bytes long.
; These are small blocks -- residuals, or to handle boundary/alignment cases.
;
; The amount of data to process (bsize) must be a non-zero multiple of 4.
;
;-----------------------------------------------------------------------------

Asm_DoBlock4 proc uses xbx xsi xdi xbp, \
	inputbuffer:ptr byte, outputbuffer:ptr byte, \
	bsize:size_t, pMtab:ptr dword

	mov	xbx,if64(r8, bsize)	;bytes to process (multiple of 4)
	mov	xsi,if64(rcx, inputbuffer)  ;source
	mov	xdi,if64(rdx, outputbuffer) ;destination
	mov	xbp,if64(r9, pMtab)	;combined multiplication table
lp4:
	mov	eax,[xsi]
	add	xsi,4
	mov	ecx,eax
	shr	eax,16
	movzx	edx,al
	shr	eax,8
	mov	edx,[xbp+xdx*4]		;look up 2nd solution
	xor	edx,[xbp+400h+xax*4]
	movzx	eax,ch
	movzx	ecx,cl
	shl	edx,16
	mov	eax,[xbp+400h+xax*4]	;1st solution
	xor	eax,[xbp+xcx*4]
	xor	edx,eax			;combined solution
	xor	edx,[xdi]
	mov	[xdi],edx
	add	xdi,4
	sub	xbx,4			;adjust src/dst count, zero if done
	jnz	lp4

	ret
Asm_DoBlock4 endp

;-----------------------------------------------------------------------------
;
; Processes block of data a multiple of 8 bytes long using SIMD (mmx) opcodes.
;
; It is the caller's responsibility to insure the in/out pointers are
; optimally aligned; if not, performance is reduced significantly.
;
; The amount of data to process (bsize) must be a non-zero multiple of 8.
;
; Unfortunately this routine "reads ahead" 8 bytes beyond the true end of
; the given source block.  The caller must compensate for this by
; appropriately limiting the size of the source.
;
;-----------------------------------------------------------------------------

Asm_DoBlock8 proc uses xbx xsi xdi xbp, \
	inputbuffer:ptr byte, outputbuffer:ptr byte, \
	bsize:size_t, pMtab:ptr dword

	mov	xbx,if64(r8, bsize)	;bytes to process (multiple of 8)
	mov	xsi,if64(rcx, inputbuffer)  ;source
	mov	xdi,if64(rdx, outputbuffer) ;destination
	mov	xbp,if64(r9, pMtab)	;combined multiplication table

	mov	eax,[xsi]		;load 1st 8 source bytes
	movd	mm4,dword ptr [xsi+4]

	add	xsi,xbx			;point to end of input/output
	add	xdi,xbx
	neg	xbx			;convert byte size to count-up

	; The below loop processes 8 bytes at a time.  It runs at approx.
	; 10.5 clocks per iteration on a Core 2 Duo.
	; That's close to optimal -- which is 10 -- since there are 10 mem
	; reads in the loop (there's only one mem read port in the
	; architecture).  The extra .5 clock appears to be a
	; cache-loading delay, since if the read addresses below are made
	; invariant it goes away.
	;
	; The 8 byte MMX loop runs 20-30% faster than a simple (non-MMX) four
	; byte loop on a Core 2 Duo.  The speedup is really only due to
	; it doing wider memory fetches (8 bytes per src/dst instead of 4).
	;
	; Some work was done to achieve U/V pipe pairing for early Pentiums,
	; but not at the expense of the Core 2 Duo performance.

	align	16
lp8:
	movzx	edx,al
	movzx	ecx,ah
	shr	eax,16
	movd	mm0,dword ptr [xbp+xdx*4]
	movd	mm1,dword ptr [xbp+400h+xcx*4]
	movzx	edx,al
	movzx	ecx,ah
	movd	eax,mm4
	movq	mm4,[xsi+xbx+8]		;read-ahead next 8 source bytes
	movd	mm2,dword ptr [xbp+xdx*4]
	movzx	edx,al
	movq	mm5,[xdi+xbx]
	movd	mm3,dword ptr [xbp+400h+xcx*4]
	movzx	ecx,ah
	shr	eax,16
	punpckldq mm0,[xbp+xdx*4]
	movzx	edx,al
	punpckldq mm1,[xbp+400h+xcx*4]
	movzx	ecx,ah
	punpckldq mm2,[xbp+xdx*4]
	pxor	mm1,mm0
	punpckldq mm3,[xbp+400h+xcx*4]
	pxor	mm3,mm2
	pslld	mm3,16
	movd	eax,mm4			;prepare src bytes 3-0 for next loop
	pxor	mm1,mm5
	psrlq	mm4,32			;align src bytes 7-4 for next loop
	pxor	mm1,mm3
	movq	[xdi+xbx],mm1
	add	xbx,8
	jnz	lp8

	ret
Asm_DoBlock8 endp

	end
